package com.timeseries;

import java.util.Iterator;
import java.util.List;

public class TimeSeriesDouble extends TimeSeries<Double> {

	public TimeSeriesDouble(){
		super();
	}
	
	public TimeSeriesDouble(int numOfDimensions){
		super(numOfDimensions);
	}
	
	public TimeSeriesDouble(String string, boolean b, boolean c, char d) {
		super(string, b, c, d);
	}

	@Override
	protected TimeSeriesPoint<Double> getTimeSeriesPoint(List<String> subList) {
		Double[] values = new Double[subList.size()];
		Iterator<String> iter = subList.iterator();
		int i = 0;
		while (iter.hasNext()) {
			String val = iter.next();
			// Attempt to parse the next value to a Double value.
			Double nextValue;
			try {
				nextValue = Double.valueOf(val);
			} catch (NumberFormatException e) {
				throw new InternalError("ERROR:  '" + val
						+ "' is not a valid number");
			}
			values[i++] = nextValue;

		}
		return new TimeSeriesPoint<>(values);
	}

	@Override
	public void normalize() {
		// Calculate the mean of each FD.
		final double[] mean = new double[this.numOfDimensions()];
		for (int col = 0; col < numOfDimensions(); col++) {
			double currentSum = 0.0;
			for (int row = 0; row < this.size(); row++)
				currentSum += this.getMeasurement(row, col);

			mean[col] = currentSum / this.size();
		} // end for loop

		// Calculate the standard deviation of each FD.
		final double[] stdDev = new double[numOfDimensions()];
		for (int col = 0; col < numOfDimensions(); col++) {
			double variance = 0.0;
			for (int row = 0; row < this.size(); row++)
				variance += Math.abs(getMeasurement(row, col) - mean[col]);

			stdDev[col] = variance / this.size();
		} // end for loop

		// Normalize the values in the data using the mean and standard
		// deviation
		// for each FD. => Xrc = (Xrc-Mc)/SDc
		for (int row = 0; row < this.size(); row++) {
			for (int col = 0; col < numOfDimensions(); col++) {
				// Normalize data point.
				if (stdDev[col] == 0.0) // prevent divide by zero errors
					setMeasurement(row, col, 0.0); // stdDev is zero means all
													// pts identical
				else
					// typical case
					setMeasurement(row, col,
							(getMeasurement(row, col) - mean[col])
									/ stdDev[col]);
			} // end for loop
		} // end for loop
	} // end normalize();

}
