/*
 * TimeSeriesPoint.java   Jul 14, 2004
 *
 * Copyright (c) 2004 Stan Salvador
 * stansalvador@hotmail.com
 */

package com.timeseries;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.math.BigInteger;


public class TimeSeriesPoint<T>
{
   // PRIVATE DATA
   private T[] measurements;
   private int hashCode;


   // CONSTRUCTORS
   public TimeSeriesPoint(T[] values)
   {
      hashCode = 0;
      measurements = values.clone();
      for (int x=0; x<values.length; x++)
      {
         hashCode += values[x].hashCode();
         measurements[x] = values[x];
      }
   }


   public TimeSeriesPoint(Collection<T> values)
   {
      measurements = (T[]) values.toArray();
      hashCode = 0;

      final Iterator<T> i = values.iterator();
      int index = 0;
      while (i.hasNext())
      {
         final T nextElement = i.next();
         hashCode += nextElement.hashCode();
      }  // end while loop
   }  // end constructor



   // FUNCTIONS
   public T get(int dimension)
   {
      return measurements[dimension];
   }


   public void set(int dimension, T newValue)
   {
      hashCode -= measurements[dimension].hashCode();
      measurements[dimension] = newValue;
      hashCode += newValue.hashCode();
   }


   public T[] toArray()
   {
      return measurements;
   }


   public int size()
   {
      return measurements.length;
   }


   public String toString()
   {
      String outStr = "(";
      for (int x=0; x<measurements.length; x++)
      {
         outStr += measurements[x];
         if (x < measurements.length-1)
            outStr += ",";
      }
      outStr += ")";

      return outStr;
   }  // end toString()


   public boolean equals(Object o)
   {
      if (this == o)
         return true;
      else if (o instanceof TimeSeriesPoint)
      {
         final T[] testValues = (T[]) ((TimeSeriesPoint)o).toArray();
         if (testValues.length == measurements.length)
         {
            for (int x=0; x<measurements.length; x++)
               if (measurements[x] != testValues[x])
                  return false;

            return true;
         }
         else
            return false;
      }
      else
         return false;
   }  // end public boolean equals


   public int hashCode()
   {
      return this.hashCode;
   }

}  // end class TimeSeriesPoint
